/*
 * $Id: read.c,v 1.5 1998/04/07 02:42:59 labovit Exp $
 */

#include <sys/types.h> 
#include <sys/stat.h>
#include <fcntl.h>
#include <mrt.h>
#include <select.h>
#include <sys/time.h>
#include <string.h>
#include "salserver.h"


extern trace_t			*default_trace;
extern salserver_t		SALSERVER;

int more_to_write (int fd) {
  fd_set fdvar_write;
  struct timeval tv;

  tv.tv_sec = 0; /* XXX */
  tv.tv_usec = 0;

  if (fd <= 0) {return 0;}

  FD_ZERO (&fdvar_write);
  FD_SET(fd, &fdvar_write);
  if (select (FD_SETSIZE, NULL, &fdvar_write, NULL, &tv) <= 0)
    return (0);

  /* server isn't ready for us -- we have timed out */
  if (FD_ISSET (fd, &fdvar_write))
    return (1); 

  return (0);
}

int more_to_read (int fd) {
  fd_set fdvar_read;
  struct timeval tv;

  tv.tv_sec = 0; /* XXX */
  tv.tv_usec = 0;

  FD_ZERO (&fdvar_read);
  FD_SET(fd, &fdvar_read);
  select (FD_SETSIZE, &fdvar_read, NULL, NULL, &tv);

  /* server isn't ready for us -- we have timed out */
  if (FD_ISSET (fd, &fdvar_read))
    return (1); 

  return (0);
}


void reset_data_pdu (sal_connection_t *connection) {
  data_pdu_t *data;
  
  data = connection->data_pdu;

  /* clean up memory */
  LL_Destroy (data->ll_properties);
  if (data->propbuf != NULL) {
    Delete (data->propbuf);
    data->propbuf = NULL;
  }
  if (data->databuf != NULL) {
    Delete (data->databuf);
    data->databuf = NULL;
  }
  if (data->name != NULL) {
    Delete (data->name);
    data->name = NULL;
  }

  data->time_closest = 0;
  data->cp = data->header;
  data->ll_properties = LL_Create (LL_DestroyFunction, Delete_Property, NULL);

  data->query_id = 0;
  data->state = 0; /* need header*/
}



void flush_connection (sal_connection_t *connection) {
  /* while (1) read socket  */
  reset_data_pdu (connection);
  return;
}


/* process_properties
 * We have read in at least the Salamander packet header.
 * Now, we have to parse the header to find fields, including the
 * well-hidden OBJ size field.
 * 
 * We fill in the connection's data_pdu information.
 */
int process_properties (sal_connection_t *connection) {
  data_pdu_t *data;
  sal_property_t *sal_property;
  char *key, *value, *end, *cp;

  data = connection->data_pdu;

  /* clean up */
  if (data->date != NULL) {
    Delete (data->date);
    data->date = NULL;
  }

  sal_property = New (sal_property_t);
  end = (char *) data->cp;
  cp = (char *) data->propbuf;

  for (key = cp; *key && key < end;) {
    value = key + strlen(key) + 1;
		
    /* Invalid plist: */
    if (value >= end) {
      trace (NORM, default_trace, "Invalid property list!\n");
      exit;
    }

    sal_property->key = strdup (key);
    sal_property->value = strdup (value);
    LL_Add (data->ll_properties, sal_property);
    trace (TR_TRACE, default_trace, "Recv from %s %s:%s \n", connection->name, key, value);

    if (strcasecmp ("OBJ_SIZE", sal_property->key) == 0) {
      data->datalen =  atol (sal_property->value);
    }
    else if (strcasecmp ("NAME", sal_property->key) == 0) {
      data->name = strdup (sal_property->value);
    }
    else if (strcasecmp ("DATE", sal_property->key) == 0) {
      data->date = strdup (sal_property->value);
    }
    else if (strcasecmp ("TIME", sal_property->key) == 0) {
      data->time  = atol (sal_property->value);
    }
    else if (strcasecmp ("TIME_NEXT", sal_property->key) == 0) {
      data->time_closest = atol (sal_property->value);
    }
    else if (strcasecmp ("ID", sal_property->key) == 0) {
      data->query_id  = atol (sal_property->value);
    }
    else if (strcasecmp ("COMMAND", sal_property->key) == 0) {
      /* publish */
      if (strcasecmp ("PUBLISH", sal_property->value) == 0)
	data->flag = SAL_COMMAND_PUBLISH;  
      /* meta-query */
      else if (strcasecmp ("META-QUERY", sal_property->value) == 0)
	data->flag = SAL_COMMAND_META_QUERY;  
      /* query */
      else if (strcasecmp ("UNSUBSCRIBE", sal_property->value) == 0)
	data->flag = SAL_COMMAND_UNSUBSCRIBE;
      else
	data->flag = SAL_COMMAND_QUERY;  
    }
    
    sal_property = New (sal_property_t);
    key = (char *)(value + strlen(value) + 1);
  }

  Delete (sal_property); /* always delete last property -- not used */

  if (data->datalen> SALSERVER.max_data_size) {
    trace (NORM | TR_ERROR, default_trace, "ERROR -- %d bytes greater %d\n",
	   data->datalen, SALSERVER.max_data_size);
    sal_delete_connection (connection);
    return (-1);
  }

  data->databuf = malloc (data->datalen);
  data->cp = data->databuf;

  data->state = 2; /* need data (or maybe not) */
  return (1);
}

void process_header (sal_connection_t *connection) {
  data_pdu_t *data;
  u_long ckey, version, plist_len;
  char *cp;

  if (connection->state != 1) {
    return;
  }

  data = connection->data_pdu;
  cp = (char *) data->header;

  UTIL_GET_NETLONG (ckey, cp);
  UTIL_GET_NETLONG (version, cp);
  UTIL_GET_NETLONG (plist_len, cp);

  data->proplen = plist_len;
  trace (TR_TRACE, default_trace, "-- Allocating %d bytes for proplen...\n", plist_len);

  data->propbuf = malloc (plist_len);
  data->cp = data->propbuf;

  data->state = 1; /* need properites */
  return;
}


void sal_read (sal_connection_t *connection) {
  data_pdu_t *data; 
  int bytes_read;
  int n;

  /* check if scheduled for deletion */
  if (connection->state == 3) {
    return;
  }
  
  /* get key */
  if (connection->state == 0) {
    char buf[12], *cp;
    int key;
    connection->state = 1;
    
    if (read (connection->sockfd, buf, 4) <= 0) {
      trace (NORM |TR_ERROR, default_trace, "Read of key failed...\n");
      select_delete_fd (connection->sockfd);
      /* delete connection memory */
      return;
    }
    cp = buf;
    UTIL_GET_NETLONG (key, cp);

    trace (NORM, default_trace, "Adding connection for %s (fd = %d, key = %d)\n",
	   connection->name, connection->sockfd, key);
    LL_Add (SALSERVER.ll_sal_connections, connection);
    connection->data_pdu->state = 0;
    connection->data_pdu->cp = connection->data_pdu->header;
    if (!more_to_read (connection->sockfd)) {
      select_enable_fd_mask (connection->sockfd, SELECT_READ);
      return;
    }
  }

  data = connection->data_pdu;

  /* need header */
  if (data->state == 0) {
    bytes_read = data->cp - data->header;

    if ((n = read (connection->sockfd, data->cp, 12 - bytes_read)) <= 0) {
      sal_delete_connection (connection);
      return;
    }
   
    data->cp += n;
    if (data->cp - data->header == 12) {
      process_header (connection);

      if (!more_to_read (connection->sockfd)) {
	select_enable_fd_mask (connection->sockfd, SELECT_READ);
	return;
      }
    }
    else {
      select_enable_fd_mask (connection->sockfd, SELECT_READ);
      return;
    }
  }


  /* need properties */
  if (data->state == 1) {
    bytes_read = data->cp - data->propbuf;

    if ((n = read (connection->sockfd, data->cp, data->proplen - bytes_read)) <= 0) {
      sal_delete_connection (connection);
      return;
    }
    
    data->cp += n;
    /* we have all the properties in buffer */
    if ( (data->cp - data->propbuf) == data->proplen) {
      
      if (process_properties (connection) < 0) {
	sal_delete_connection (connection);
	return;
      }

      if ((data->flag == SAL_COMMAND_QUERY) || 
	  (data->flag == SAL_COMMAND_META_QUERY)) {

    	n = sal_process_query (connection, connection->data_pdu);
	/* if (n < 0) return; error */
	if (n == 1) /* save the query -- we can't touch this memory */
	  connection->data_pdu = (data_pdu_t *) New_Data_Pdu ();

	select_enable_fd_mask (connection->sockfd, SELECT_READ);
	reset_data_pdu (connection);
	return;
      }
      if (data->flag == SAL_COMMAND_UNSUBSCRIBE) {
	sal_process_unsubscribe (connection, connection->data_pdu);
	select_enable_fd_mask (connection->sockfd, SELECT_READ);
	reset_data_pdu (connection);
	return;
      }

	

      if (!more_to_read (connection->sockfd)) {
	select_enable_fd_mask (connection->sockfd, SELECT_READ);
	return;
      }
    }
    /* properites not all read in yet */
    else {
      select_enable_fd_mask (connection->sockfd, SELECT_READ);
      return;
    }
  }

  /* need data */
  if (data->state == 2) {
    bytes_read = data->cp - data->databuf;

    if (data->datalen <= 0) {
      trace (NORM, default_trace, "Error -- %d byes published!\n", data->datalen);    
    }

    if ((n = read (connection->sockfd, data->cp, data->datalen - bytes_read)) <= 0) {
      sal_delete_connection (connection);
      return;
    }

    select_enable_fd_mask (connection->sockfd, SELECT_READ);

    data->cp += n;

    if ((data->cp - data->databuf) == data->datalen) {
      sal_process_data (connection);
      reset_data_pdu (connection);
      return;
    }
    return;
  }
}




data_pdu_t *sal_read_disk (int fd, char *source) {
  char htmp[12], *buf, *cp, *end;
  u_long ckey, version, plist_len;
  data_pdu_t	*data_pdu;
  char *key, *value;
  sal_property_t *sal_property;
  int datalen;

  if (read (fd, htmp, 12) < 1) {
    trace (NORM, default_trace, "Read header from %s FAILED!\n", source);
    return (NULL);
  }
  cp = htmp;

  UTIL_GET_NETLONG (ckey, cp);
  UTIL_GET_NETLONG (version, cp);
  UTIL_GET_NETLONG (plist_len, cp);

  if (plist_len > SALSERVER.max_data_size) {
    trace (NORM, default_trace, "ERROR -- %d bytes greater %d\n",
	   plist_len, SALSERVER.max_data_size);
    /* flush socket */
    return (NULL);
  }

  if ((buf = malloc (plist_len)) == NULL) {
    trace (NORM | TR_ERROR, default_trace, 
	   "ERROR -- malloc for %d bytes failed sal_connection_read\n",
	   plist_len);
    return (NULL);
  }

  /* read properties */
  if (read (fd, buf, plist_len) < 1) {
    trace (NORM, default_trace, "Read properties from %s FAILED!\n", source);
    Delete (buf);
    return (NULL);
  }
  cp = buf;
  end = cp + plist_len;

  data_pdu = (data_pdu_t *) New_Data_Pdu ();

  /* store property buffer */
  data_pdu->proplen = plist_len;
  if ((data_pdu->propbuf = malloc (plist_len)) == NULL) {
    trace (NORM | TR_ERROR, default_trace, 
	   "ERROR -- malloc for %d bytes failed sal_connection_read\n",
	   plist_len);
    return (NULL);
  }
  memcpy (data_pdu->propbuf, cp, plist_len);

  sal_property = New (sal_property_t);
  datalen = 0;

  for (key = cp; *key && key < end;) {
    value = key + strlen(key) + 1;
		
    /* Invalid plist: */
    if (value >= end) {
      trace (NORM, default_trace, "Invalid property list!\n");
      exit;
    }

    sal_property->key = strdup (key);
    sal_property->value = strdup (value);
    LL_Add (data_pdu->ll_properties, sal_property);

    if (strcmp ("OBJ_SIZE", sal_property->key) == 0) {
      datalen =  atol (sal_property->value);
    }
    if (strcmp ("NAME", sal_property->key) == 0) {
      data_pdu->name = strdup (sal_property->value);
    }
    if (strcmp ("DATE", sal_property->key) == 0) {
      data_pdu->date = strdup (sal_property->value);
    }
    if (strcmp ("TIME", sal_property->key) == 0) {
      data_pdu->time = atol (sal_property->value);
    }
    if (strcmp ("COMMAND", sal_property->key) == 0) {
      if (strcmp ("PUBLISH", sal_property->value) == 0)
	data_pdu->flag = SAL_COMMAND_PUBLISH;  /* publish */
      else
	data_pdu->flag = SAL_COMMAND_QUERY;  /* query */
    }
    
    sal_property = New (sal_property_t);
    key = (char *)(value + strlen(value) + 1);
  }

  Delete (sal_property); /* always delete last property -- not used */

  if (plist_len > SALSERVER.max_data_size) {
    trace (NORM | TR_ERROR, default_trace, "ERROR -- %d bytes greater %d\n",
	   plist_len, SALSERVER.max_data_size);
    /* flush socket */
    return (NULL);
  }

  /* make sure we have a channel name -- this is the one and only
   * mandatory field !
   */
  if (data_pdu->name == NULL) {
    trace (NORM | TR_ERROR, default_trace, "ERROR -- no channel name from %s\n", source);
    Delete_Data (data_pdu);
    return (NULL);
  }

  Delete (buf); /* free up memory used by properties */
  data_pdu->datalen = datalen;
  
  if (datalen > 0) {
    buf = malloc (datalen); /* memory for data */
    trace (TR_TRACE, default_trace, "Loaded %s (%d bytes)\n",
	   source, datalen);
    if (read (fd, buf, datalen) != datalen) {
      trace (NORM, default_trace, "Error! Loading data\n");
      return (NULL);
    }
    data_pdu->databuf = malloc (datalen);  
    memcpy (data_pdu->databuf, buf, datalen);
    Delete (buf);
  }

  

  return (data_pdu);
}
